unit Storage;

{        Takes care of all storage requirements in the LISP interpreter}
{        Includes these things :
                 Heap for dynamic variable storage
                Storing of data types on the program list
}
interface
uses Declars, Envs,Errors;

var TokenList      :    CellPtr; {Pointer to the list of tokens}
    FreeCells      :    CellPtr; {Stack of inaccessible holes in heap}

{Procedure: CreateCons
 Description: Creates a construct from two cells
 Arguments: The two cells to contruct from and the variable to store result
 Returns:NA}
procedure CreateCons(var Token: cell; Car: cell; Cdr:cell);

{Procedure:CreateSym
 Description:Create a token of type Cell_Type
 Arguments:Name,and type of the cell and the variable to store result
 Returns:NA}
procedure CreateSym( var Token : cell; Name : string; Cell_Type :DataType);

{Procedure:CreateNum
 Description:Create a number atom
 Arguments:Name,and value of the cell and the variable to store result
 Returns:NA}
procedure CreateNum( var Token:cell; Name:string; Number: real);

{Procedure:CreateStr
 Description:Create a string atom
 Arguments:Name,and value of the cell and the variable to store result
 Returns:NA}
procedure CreateStr( var Token:cell; Name:string; Str: string);

{Procedure:CreateBool
 Description:Create a bool atom
 Arguments:Name,and value of the cell and the variable to store result
 Returns:NA}
procedure CreateBool( var Token:cell; Name:string; Bool: boolean);

{Garbage collection related procedures}

{Procedure:C_New
 Description: C_New is similar to the Pascal new function but does not leave holes behind
 Arguments:The pointer to create to, the contents of the new variable
 Returns:NA}
procedure C_New(var CellPointer:CellPtr; Contents:Cell);

{Procedure:C_Dispose
 Description: C_Dispose is similar to the Pascal dispose function but does not leave holes behind
 Arguments: The pointer to create to, the contents of the new variable
 Returns:NA}
procedure C_Dispose(var CellPointer:CellPtr);

{Procedure: MarkCell
 Description:Protects useful cells from being garbage collected
 Arguments: A pointer to the list to mark, the value to mark it as
 Returns: NA}
procedure MarkCell(var Node:CellPtr;Rubbish:boolean);

{Procedure:GarbageCollect
 Description: Variation of a mark - sweep garbage collection
 Arguments:NA
 Returns:NA}
procedure GarbageCollect;

{Procedure:LinkToken
 Description:Every cell is added into a linked list which is used during garbage collection
 Arguments:The cell to add
 Returns:NA}
procedure LinkToken(Node: CellPtr);

{Procedure:DisposeCons
 Description: Dispose a cons - straight off
 Arguments:  The head of the cons
 Returns:NA}
procedure DisposeCons(Node: CellPtr);

{Procedure:InitGlobalEnv
 Description:Initialises env table
 Arguments:NA
 Returns:NA}
procedure InitGlobalEnv;

implementation
{---------------------------------------------------------------------------}
procedure CreateSym( var Token : cell; Name : string; Cell_Type :DataType);
begin
     Token := _NULL;
     Token.Name := Name;
     Token.Cell_Type := Cell_Type;
end;
{---------------------------------------------------------------------------}
procedure CreateNum( var Token:cell; Name:string; Number: real);
begin
     Token:=_NULL;
     Token.Name:=Name;
     Token.Cell_Type:=tNum;
     Token.value_Num := Number;
end;
{---------------------------------------------------------------------------}
procedure CreateBool( var Token:cell; Name:string; Bool: boolean);
begin
     Token:=_NULL;
     Token.Name:=Name;
     Token.Cell_Type:=tBool;
     Token.value_Bool := Bool;
end;
{---------------------------------------------------------------------------}
procedure CreateStr( var Token:cell; Name:string; Str: string);
begin
     Token:=_NULL;
     Token.Name:=Name;
     Token.Cell_Type:=tString;
     Token.value_Str := Str;
end;
{---------------------------------------------------------------------------}
procedure CreateCons(var Token: cell; Car: cell; Cdr:cell);
begin
      C_New(Token.car,Car);  {Create head}
      C_New(Token.cdr,Cdr);  {Create Rest}

     Token.car^.Rubbish:=true;
     Token.cdr^.Rubbish:=true;

     LinkToken(Token.car);   {Add to tokenlist}
     LinkToken(Token.cdr);

     Token.Cell_Type := tCons;
end;
{---------------------------------------------------------------------------}
procedure LinkToken(Node: CellPtr);
begin
       {insert the new node into the tokenlist}
       Node^.ProtPtr := TokenList;
       TokenList := Node;
end;
{---------------------------------------------------------------------------}
procedure DisposeCons(Node: CellPtr);
var List : CellPtr;
begin
{Recursively move to the end of the cons}
    if Node^.cdr <> nil then DisposeCons(Node^.cdr);
    if Node^.car <> nil then DisposeCons(Node^.car);

{Remove cell from tokenlist to prevent broken list}
  List:=TokenList;

   while ((List<>nil) and (List^.Protptr<>Node)) do
     List:=List^.Protptr;

  if TokenList=Node then TokenList:=Node^.Protptr  {If head of list}
  else if List <> nil then List^.Protptr:=Node^.Protptr; {otherwise skip it}

  C_Dispose(Node);     {Delete the cell}
end;
{---------------------------------------------------------------------------}
procedure MarkCell(var Node:CellPtr;Rubbish:boolean);
begin
    Node^.Rubbish:=Rubbish; {Recursively mark each node in the list}
    if Node^.car <> nil then MarkCell(Node^.car,Rubbish);
    if Node^.value <> nil then MarkCell(Node^.value,Rubbish);
    if Node^.HashPtr <> nil then MarkCell(Node^.HashPtr,Rubbish);
    if Node^.cdr <> nil then MarkCell(Node^.cdr,Rubbish);
end;
{---------------------------------------------------------------------------}
procedure GarbageCollect;

{Cleans the useless pointers from the external chains of the htable}
procedure UprootHashChains(Node:CellPtr);
begin
     if Node^.Hashptr<>nil then UprootHashChains(Node^.Hashptr);
     if ((Node^.Hashptr<> nil) and
         (Node^.Hashptr^.Rubbish=true)) then Node^.Hashptr:=nil;
end;

var P   :  CellPtr;
    i   :  integer;

begin
{Traverse Tokenlist deleting the rubbish from it}
  if TokenList <> nil then
  begin
       for i := 1 to Hashsize do
     begin
       if Environ.h_table[i]<>nil then
       begin
           if Environ.h_table[i]^.hashptr <> nil then
                 UprootHashChains(Environ.h_table[i]);

          Environ.h_table[i]^.next:=nil;

          if Environ.h_table[i]^.Rubbish=true then
                 Environ.h_table[i]:=nil;
       end;
     end;

     repeat
       P:= TokenList;
       TokenList:=TokenList^.Protptr;
       if P^.Rubbish = true then C_Dispose(P);
     until TokenList= nil;

{Clean up useless base links in htable}
   end;
end;
{---------------------------------------------------------------------------}
 {Modified new, dispose functions}
{The C_New algorithm checks if there is anything on the freecell stack}
{If so pop off the top and store Contents there}
{Otherwise create a cell in contiguous memory}
procedure C_New(var CellPointer: CellPtr; Contents:Cell);
var Temp            :         Pointer;

begin
{$IFNDEF DPMI}
     if FreeCells<> nil then  {If there are some holes...}
      begin
        CellPointer:= FreeCells;  {write data into hole}
        FreeCells:=FreeCells^.Protptr;
     end
     else begin
{$ENDIF}
      new(CellPointer); {Else continue taking up memory contiguously}
{$IFNDEF DPMI}
    end;
{$ENDIF}
     CellPointer^:=Contents;
     CellPointer^.Address:=CellPointer;
end;
{---------------------------------------------------------------------------}
{The C_Dispose algorithm- first disposes the Cell.
 If it finds that it has left a hole.  It pushes the address of the
 hole onto the freecells stack}
{---------------------------------------------------------------------------}
procedure C_Dispose(var CellPointer:CellPtr);
var Temp            :        Pointer;
    StackNode       :        CellPtr;
begin
 dispose(CellPointer);
{$IFNDEF DPMI}
     if FreeList<>HeapPtr then {If a hole was created}
     begin
          CellPointer^:=_NULL;
          Temp := HeapPtr;
          HeapPtr:= FreeList;
          new(StackNode);
          StackNode^.Protptr := FreeCells; {Add hole to list of free cells}
          FreeCells := StackNode;
          HeapPtr:=Temp;
          FreeList:=HeapPtr;
    end;
{$ENDIF}
end;
{-------------------------------------------------------------------------}
procedure InitGlobalEnv;
var
    Token       :       cell;
    i           :       integer;
begin
    Level:=0;   {Set to global enviroment}
    CurLevel:=0;
    Environ.Initialise; {Set environment to nil}

    for i:= 1 to Num_defined do
    begin
         CreateSym(Token,Predefined[i],tPredefined);
         Environ.Add(Token,Predefined[i]);
    end;
    Environ.Add(_TRUE,'T'); {Any constants should be placed here}
    Environ.Add(_NULL,'NIL');
end;
{---------------------------------------------------------------------------}
end.